/*
** Module   : library_linux.c
** Abstract : implementation of operating-system specific helpers for library management
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 GES 20131024 First version.
*/ 

#include "library.h"
#include <dlfcn.h>

// local function declarations
void*    int64ToPointer(int64_t modhandle);
int64_t  pointerToInt64(void* ptr);

/**
 * Attempts to load the named library.  This will use the operating-system runtime library
 * loading mechanism.
 *
 * @param    libname
 *           This may be any valid library filename for the current system, including names
 *           that contain path separators (both relative and absolute). This should not be
 *           <code>null</code> or the empty string.
 *
 * @return   An operating-system specific "handle" to the library. Under Windows this is
 *           the "module handle" of type HMODULE and under Linux it is just referred to as
 *           a "handle" that is returned as a void pointer.  In either case, if
 *           <code>NULL</code> (0) is returned, then the call failed.
 */ 
int64_t load_library_native(const char* libname)
{
   // the use of RTLD_LAZY as the flags value is just a guess, based on behavior it is possible
   // that Progress uses RTLD_NODELETE as well
   void* ptr = dlopen(libname, RTLD_LAZY | RTLD_NODELETE);
   return pointerToInt64(ptr);
}

/**
 * Attempts to unload the library identified by the "handle" which was returned from a
 * previous call to load_library_native.  This will use the operating-system runtime library
 * unloading mechanism. Although this call may fail, Progress seems to always silently ignore
 * any failures, so this method has no return value.
 *
 * @param    modhandle
 *           This must be a valid handle to the library.
 */ 
void unload_library_native(int64_t modhandle)
{
   dlclose(int64ToPointer(modhandle));
}

/**
 * Uses the operating-system runtime library lookup facility to obtain the address of the
 * given symbol in the library.  The symbol name must match exactly, including the case
 * and any decorations that exist as exported in the library.
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 * @param    funcname
 *           This is the exact symbol to lookup in the library.
 *
 * @return   The memory address of the function or <code>NULL</code> (0) if the function cannot
 *           be found by that name in the given library.
 */ 
int64_t find_by_name_native(int64_t modhandle, const char* funcname)
{
   void* addr = dlsym(int64ToPointer(modhandle), funcname);
   return pointerToInt64(addr);
}

/**
 * Uses the operating-system runtime library lookup facility to obtain the address of the
 * given ordinal in the library.  Only some operating systems allow exporting functions by
 * ordinal number.  Those that do allow it, will typically have a table of contiguous ordinal
 * values in the library, each of which represents a function that may or may not be also
 * exported by name (as a symbol).
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 * @param    ordinal
 *           The ordinal to lookup in the library.  Only the lower word is honored.
 *
 * @return   Linux/UNIX systems don't support ordinal exports. This always returns 0.
 */ 
int64_t find_by_ordinal_native(int64_t modhandle, uint16_t ordinal)
{
   return 0;
}

/**
 * Convert the given integer value into a void pointer in a manner that is safe on either
 * 32-bit or 64-bit systems.
 *
 * @param    modhandle
 *           This must be a valid handle to a library that is already loaded.
 *
 * @return   The module handle as a void pointer.
 */
void* int64ToPointer(int64_t modhandle)
{
   #ifdef WORD_SIZE_64
   void* handle = (void*) modhandle;
   #else
   void* handle = (void*) ((int32_t) modhandle);
   #endif
   
   return handle;
}

/**
 * Convert the given void pointer into an integer value in a manner that is safe on either
 * 32-bit or 64-bit systems.
 *
 * @param    ptr
 *           This must be a pointer.
 *
 * @return   The module handle as an integer.
 */
int64_t pointerToInt64(void* ptr)
{
   #ifdef WORD_SIZE_64
   int64_t result = (int64_t) ptr;
   #else
   int32_t result = (int32_t) ptr;
   #endif
   
   return (int64_t) result;
}
